@@ -1,4 +1,3 @@
-/* $Id: udpkg.c 64419 2010-08-20 22:27:14Z jkoenig-guest $ */
 #include "udpkg.h"
 
 #include <errno.h>
@@ -133,16 +132,62 @@ static int dpkg_dounpack(struct package_t *pkg)
 		"templates", "menutest", "isinstallable",
 		"config"
 	};
+	FILE *infp = NULL;
+	const char *compression_type = NULL;
+	const char *decompression_tool;
 #ifdef DOREMOVE
 	char *p;
-	FILE *infp = NULL, *outfp = NULL;
+	FILE *outfp = NULL;
 #endif
 
 	DPRINTF("Unpacking %s\n", pkg->package);
 
 	cwd = getcwd(0, 0);
 	chdir("/");
-	snprintf(buf, sizeof(buf), "ar -p %s data.tar.gz|tar -xzf -", pkg->file);
+
+	snprintf(buf, sizeof(buf), "ar -t %s", pkg->file);
+	if ((infp = popen(buf, "r")) == NULL)
+	{
+		FPRINTF(stderr, "Cannot retrieve archive members of %s: %s\n",
+			pkg->file, strerror(errno));
+		r = 1;
+		goto reset_cwd;
+	}
+
+	while (fgets(buf, sizeof(buf), infp)) {
+		if (strncmp(buf, "data.tar.", 9) == 0) {
+			compression_type = buf + 9;
+			break;
+		}
+	}
+	pclose(infp);
+
+	if (compression_type == NULL) {
+		FPRINTF(stderr, "No data member found in %s\n", pkg->file);
+		r = 1;
+		goto reset_cwd;
+	}
+
+	if (strcmp(compression_type, "gz\n") == 0)
+	{
+		compression_type = "gz";
+		decompression_tool = "gunzip";
+	}
+	else if (strcmp(compression_type, "xz\n") == 0)
+	{
+		compression_type = "xz";
+		decompression_tool = "unxz";
+	}
+	else
+	{
+		FPRINTF(stderr, "Invalid compression type for data member of %s\n",
+			pkg->file);
+		r = 1;
+		goto reset_cwd;
+	}
+
+	snprintf(buf, sizeof(buf), "ar -p %s data.tar.%s|%s -c|tar -x",
+		pkg->file, compression_type, decompression_tool);
 	if ((r = di_exec_shell_log(buf)) == 0)
 	{
 		/* Installs the package scripts into the info directory */
@@ -205,8 +250,8 @@ static int dpkg_dounpack(struct package_t *pkg)
 		 * so oddly...
 		 */
 		snprintf(buf, sizeof(buf),
-			"ar -p %s data.tar.gz|tar -tzf -",
-			pkg->file);
+			"ar -p %s data.tar.%s|%s -c|tar -t",
+			pkg->file, compression_type, decompression_tool);
 		snprintf(buf2, sizeof(buf2),
 			"%s%s.list", INFODIR, pkg->package);
 		if ((infp = popen(buf, "r")) == NULL ||
@@ -235,7 +280,7 @@ static int dpkg_dounpack(struct package_t *pkg)
 				}
 				fputs(p, outfp);
 			}
-		fclose(infp);
+		pclose(infp);
 		fclose(outfp);
 #endif
 
@@ -251,6 +296,8 @@ static int dpkg_dounpack(struct package_t *pkg)
 	}
 	else
 		FPRINTF(stderr, "%s exited with status %d\n", buf, r);
+
+reset_cwd:
 	chdir(cwd);
 	return r;
 }
